@@ -25,9 +25,11 @@
 
 #include "arg.h"
 #include "object.h"
+#include "foreign.h"
 #include "boxed.h"
 #include "union.h"
 #include "value.h"
+#include "gjs/byteArray.h"
 #include <gjs/gjs.h>
 
 #include <util/log.h>
@@ -44,7 +46,7 @@ _gjs_flags_value_is_valid(JSContext   *context,
     /* FIXME: Do proper value check for flags with GType's */
     if (gtype == G_TYPE_NONE)
         return JS_TRUE;
-    
+
     klass = g_type_class_ref(gtype);
 
     /* check all bits are defined for flags.. not necessarily desired */
@@ -100,43 +102,23 @@ _gjs_enum_value_is_valid(JSContext  *context,
     return found;
 }
 
-/* Return the proper GI int type for the size and signedness. */
+/* The typelib used to have machine-independent types like
+ * GI_TYPE_TAG_LONG that had to be converted; now we only
+ * handle GType specially here.
+ */
 static inline GITypeTag
-type_tag_from_size(int intsize, gboolean is_signed)
-{
-    /* Constant folding should be able to reduce this to a single constant,
-     * given constant inputs. */
-    switch (intsize) {
-    case 1: return (is_signed) ? GI_TYPE_TAG_INT8 : GI_TYPE_TAG_UINT8;
-    case 2: return (is_signed) ? GI_TYPE_TAG_INT16 : GI_TYPE_TAG_UINT16;
-    case 4: return (is_signed) ? GI_TYPE_TAG_INT32 : GI_TYPE_TAG_UINT32;
-    case 8: return (is_signed) ? GI_TYPE_TAG_INT64 : GI_TYPE_TAG_UINT64;
-    default: g_assert_not_reached ();
-    }
-}
-
-/** Convert machine-specific integer type tags such as 'int' and 'long'
- *  into machine-independent explicitly-sized type tags such as 'int32'. */
-static GITypeTag
-normalize_int_types(GITypeTag type) {
-    enum { UNSIGNED=FALSE, SIGNED=TRUE };
-    switch (type) {
-#define INT_TYPE(tag, ty, sign) \
-    case GI_TYPE_TAG_##tag: return type_tag_from_size(sizeof(ty), (sign));
-        INT_TYPE(SHORT, short, SIGNED);
-        INT_TYPE(USHORT, unsigned short, UNSIGNED);
-        INT_TYPE(INT, int, SIGNED);
-        INT_TYPE(UINT, unsigned int, UNSIGNED);
-        INT_TYPE(LONG, long, SIGNED);
-        INT_TYPE(ULONG, unsigned long, UNSIGNED);
-        INT_TYPE(SSIZE, ssize_t, SIGNED);
-        INT_TYPE(SIZE, size_t, UNSIGNED);
-        INT_TYPE(TIME_T, time_t, SIGNED); /* time_t is signed */
-        INT_TYPE(GTYPE, GType, UNSIGNED);
-#undef INT_TYPE
-    default:
-        return type; /* not a weird int type, return untouched */
+replace_gtype(GITypeTag type) {
+    if (type == GI_TYPE_TAG_GTYPE) {
+        /* Constant folding should handle this hopefully */
+        switch (sizeof(GType)) {
+        case 1: return GI_TYPE_TAG_UINT8;
+        case 2: return GI_TYPE_TAG_UINT16;
+        case 4: return GI_TYPE_TAG_UINT32;
+        case 8: return GI_TYPE_TAG_UINT64;
+        default: g_assert_not_reached ();
+        }
     }
+    return type;
 }
 
 /* Check if an argument of the given needs to be released if we created it
@@ -371,6 +353,43 @@ gjs_object_to_g_hash(JSContext   *context,
     return JS_FALSE;
 }
 
+JSBool
+gjs_array_from_strv(JSContext   *context,
+                    jsval       *value_p,
+                    const char **strv)
+{
+    JSObject *obj;
+    jsval elem;
+    guint i;
+    JSBool result = JS_FALSE;
+
+    obj = JS_NewArrayObject(context, 0, JSVAL_NULL);
+    if (obj == NULL)
+        return JS_FALSE;
+
+    *value_p = OBJECT_TO_JSVAL(obj);
+
+    elem = JSVAL_VOID;
+    JS_AddRoot(context, &elem);
+
+    for (i = 0; strv[i] != NULL; i++) {
+        if (!gjs_string_from_utf8 (context, strv[i], -1, &elem))
+            goto out;
+
+        if (!JS_DefineElement(context, obj, i, elem,
+                              NULL, NULL, JSPROP_ENUMERATE)) {
+            goto out;
+        }
+    }
+
+    result = JS_TRUE;
+
+out:
+    JS_RemoveRoot(context, &elem);
+
+    return result;
+}
+
 JSBool
 gjs_array_to_strv(JSContext   *context,
                   jsval        array_value,
@@ -424,7 +443,7 @@ gjs_string_to_intarray(JSContext   *context,
     gsize length;
 
     element_type = g_type_info_get_tag(param_info);
-    element_type = normalize_int_types(element_type);
+    element_type = replace_gtype(element_type);
 
     switch (element_type) {
     case GI_TYPE_TAG_INT8:
@@ -520,7 +539,7 @@ gjs_array_to_array(JSContext   *context,
     GITypeTag element_type;
 
     element_type = g_type_info_get_tag(param_info);
-    element_type = normalize_int_types(element_type);
+    element_type = replace_gtype(element_type);
 
     switch (element_type) {
     case GI_TYPE_TAG_UTF8:
@@ -550,6 +569,78 @@ gjs_array_to_array(JSContext   *context,
     }
 }
 
+static JSBool
+gjs_array_to_g_array(JSContext   *context,
+                     jsval        array_value,
+                     unsigned int length,
+                     GITypeInfo  *param_info,
+                     void       **arr_p)
+{
+    GArray *array;
+    GITypeTag element_type;
+    gpointer contents;
+    guint element_size;
+
+    element_type = g_type_info_get_tag(param_info);
+    element_type = replace_gtype(element_type);
+
+    switch (element_type) {
+    case GI_TYPE_TAG_UINT8:
+    case GI_TYPE_TAG_INT8:
+      element_size = sizeof(guint8);
+      break;
+    case GI_TYPE_TAG_UINT16:
+    case GI_TYPE_TAG_INT16:
+      element_size = sizeof(guint16);
+      break;
+    case GI_TYPE_TAG_UINT32:
+    case GI_TYPE_TAG_INT32:
+      element_size = sizeof(guint32);
+      break;
+    case GI_TYPE_TAG_UINT64:
+    case GI_TYPE_TAG_INT64:
+      element_size = sizeof(guint64);
+      break;
+    default:
+        gjs_throw(context,
+                  "Unhandled GArray element-type %d", element_type);
+        return JS_FALSE;
+    }
+
+    /* create a C array */
+    if (!gjs_array_to_array (context,
+                             array_value,
+                             length,
+                             param_info,
+                             &contents))
+      return JS_FALSE;
+
+    /* append that array to the GArray */
+    array = g_array_sized_new(TRUE, TRUE, element_size, length);
+    g_array_append_vals(array, contents, length);
+
+    g_free(contents);
+
+    *arr_p = array;
+
+    return JS_TRUE;
+}
+
+static JSBool
+gjs_array_to_byte_array(JSContext  *context,
+                        jsval       value,
+                        void      **arr_p)
+{
+   GByteArray *byte_array;
+   byte_array = gjs_byte_array_get_byte_array(context,
+                                              JSVAL_TO_OBJECT(value));
+   if (!byte_array)
+       return JS_FALSE;
+
+   *arr_p = byte_array;
+   return JS_TRUE;
+}
+
 static gchar *
 get_argument_display_name(const char     *arg_name,
                           GjsArgumentType arg_type)
@@ -587,8 +678,7 @@ gjs_value_to_g_argument(JSContext      *context,
     gboolean nullable_type;
 
     type_tag = g_type_info_get_tag( (GITypeInfo*) type_info);
-    if (type_tag != GI_TYPE_TAG_TIME_T) // we handle time_t as a non-int type
-        type_tag = normalize_int_types(type_tag);
+    type_tag = replace_gtype(type_tag);
 
     gjs_debug_marshal(GJS_DEBUG_GFUNCTION,
                       "Converting jsval to GArgument %s",
@@ -679,14 +769,6 @@ gjs_value_to_g_argument(JSContext      *context,
     }
         break;
 
-    case GI_TYPE_TAG_TIME_T: {
-        double v;
-        if (!JS_ValueToNumber(context, value, &v))
-            wrong = TRUE;
-        arg->v_ulong = (unsigned long) (v/1000);
-    }
-        break;
-
     case GI_TYPE_TAG_BOOLEAN:
         if (!JS_ValueToBoolean(context, value, &arg->v_boolean))
             wrong = TRUE;
@@ -753,8 +835,13 @@ gjs_value_to_g_argument(JSContext      *context,
             interface_type = g_base_info_get_type(interface_info);
 
             switch(interface_type) {
-
             case GI_INFO_TYPE_STRUCT:
+                if (g_struct_info_is_foreign((GIStructInfo*)interface_info)) {
+                    return gjs_struct_foreign_convert_to_g_argument(
+                            context, value, type_info, arg_name,
+                            arg_type, transfer, may_be_null, arg);
+                }
+                /* fall through */
             case GI_INFO_TYPE_ENUM:
             case GI_INFO_TYPE_OBJECT:
             case GI_INFO_TYPE_INTERFACE:
@@ -765,7 +852,6 @@ gjs_value_to_g_argument(JSContext      *context,
                 gtype = g_registered_type_info_get_g_type
                     ((GIRegisteredTypeInfo*)interface_info);
                 break;
-
             case GI_INFO_TYPE_VALUE:
                 /* Special case for GValues */
                 gtype = G_TYPE_VALUE;
@@ -1042,34 +1128,52 @@ gjs_value_to_g_argument(JSContext      *context,
                 !JS_ValueToECMAUint32(context, length_value, &length)) {
                 wrong = TRUE;
             } else {
+                GIArrayType array_type = g_type_info_get_array_type(type_info);
                 GITypeInfo *param_info;
 
                 param_info = g_type_info_get_param_type(type_info, 0);
                 g_assert(param_info != NULL);
 
-                if (!gjs_array_to_array (context, value, length, param_info, &arg->v_pointer))
-                    wrong = TRUE;
+                if (array_type == GI_ARRAY_TYPE_C) {
+                    if (!gjs_array_to_array (context,
+                                             value,
+                                             length,
+                                             param_info,
+                                             &arg->v_pointer))
+                      wrong = TRUE;
+                } else if (array_type == GI_ARRAY_TYPE_ARRAY) {
+                    if (!gjs_array_to_g_array (context,
+                                               value,
+                                               length,
+                                               param_info,
+                                               &arg->v_pointer))
+                      wrong = TRUE;
+                } else if (array_type == GI_ARRAY_TYPE_BYTE_ARRAY) {
+                    if (!gjs_array_to_byte_array(context,
+                                                 value,
+                                                 &arg->v_pointer))
+                        wrong = TRUE;
+                /* FIXME: support PtrArray */
+                } else {
+                    g_assert_not_reached();
+                }
 
                 g_base_info_unref((GIBaseInfo*) param_info);
             }
         } else {
-            wrong = TRUE;
-            report_type_mismatch = TRUE;
+            GIArrayType array_type = g_type_info_get_array_type(type_info);
+            if (array_type == GI_ARRAY_TYPE_BYTE_ARRAY) {
+                if (!gjs_array_to_byte_array(context,
+                                             value,
+                                             &arg->v_pointer))
+                    wrong = TRUE;
+            } else {
+                wrong = TRUE;
+                report_type_mismatch = TRUE;
+            }
         }
         break;
 
-    case GI_TYPE_TAG_SHORT:
-    case GI_TYPE_TAG_USHORT:
-    case GI_TYPE_TAG_INT:
-    case GI_TYPE_TAG_UINT:
-    case GI_TYPE_TAG_LONG:
-    case GI_TYPE_TAG_ULONG:
-    case GI_TYPE_TAG_SIZE:
-    case GI_TYPE_TAG_SSIZE:
-    case GI_TYPE_TAG_GTYPE:
-        /* these types are converted by normalize_int_types */
-        g_assert_not_reached();
-
     default:
         gjs_debug(GJS_DEBUG_ERROR,
                   "Unhandled type %s for JavaScript to GArgument conversion",
@@ -1113,28 +1217,144 @@ gjs_value_to_g_argument(JSContext      *context,
     }
 }
 
+/* If a callback function with a return value throws, we still have
+ * to return something to C. This function defines what that something
+ * is. It basically boils down to memset(arg, 0, sizeof(*arg)), but
+ * gives as a bit more future flexibility and also will work if
+ * libffi passes us a buffer that only has room for the appropriate
+ * branch of GArgument. (Currently it appears that the return buffer
+ * has a fixed size large enough for the union of all types.)
+ */
+void
+gjs_g_argument_init_default(JSContext      *context,
+                            GITypeInfo     *type_info,
+                            GArgument      *arg)
+{
+    GITypeTag type_tag;
+
+    type_tag = g_type_info_get_tag( (GITypeInfo*) type_info);
+    type_tag = replace_gtype(type_tag);
+
+    switch (type_tag) {
+    case GI_TYPE_TAG_VOID:
+        arg->v_pointer = NULL; /* just so it isn't uninitialized */
+        break;
+
+    case GI_TYPE_TAG_INT8:
+        arg->v_int8 = 0;
+        break;
+
+    case GI_TYPE_TAG_UINT8:
+        arg->v_uint8 = 0;
+        break;
+
+    case GI_TYPE_TAG_INT16:
+        arg->v_int16 = 0;
+        break;
+
+    case GI_TYPE_TAG_UINT16:
+        arg->v_uint16 = 0;
+        break;
+
+    case GI_TYPE_TAG_INT32:
+        arg->v_int = 0;
+        break;
+
+    case GI_TYPE_TAG_UINT32:
+        arg->v_uint32 = 0;
+        break;
+
+    case GI_TYPE_TAG_INT64:
+        arg->v_int64 = 0;
+        break;
+
+    case GI_TYPE_TAG_UINT64:
+        arg->v_uint64 = 0;
+
+    case GI_TYPE_TAG_BOOLEAN:
+        arg->v_boolean = FALSE;
+        break;
+
+    case GI_TYPE_TAG_FLOAT:
+        arg->v_float = 0.0f;
+        break;
+
+    case GI_TYPE_TAG_DOUBLE:
+        arg->v_double = 0.0;
+        break;
+
+    case GI_TYPE_TAG_FILENAME:
+    case GI_TYPE_TAG_UTF8:
+    case GI_TYPE_TAG_GLIST:
+    case GI_TYPE_TAG_GSLIST:
+        arg->v_pointer = NULL;
+        break;
+
+    case GI_TYPE_TAG_INTERFACE:
+        {
+            GIBaseInfo* interface_info;
+            GIInfoType interface_type;
+
+            interface_info = g_type_info_get_interface(type_info);
+            g_assert(interface_info != NULL);
+
+            interface_type = g_base_info_get_type(interface_info);
+
+            switch(interface_type) {
+            case GI_INFO_TYPE_ENUM:
+            case GI_INFO_TYPE_FLAGS:
+                arg->v_int = 0;
+                break;
+            case GI_INFO_TYPE_VALUE:
+                /* Better to use a non-NULL value holding NULL? */
+                arg->v_pointer = NULL;
+                break;
+            default:
+                arg->v_pointer = NULL;
+                break;
+            }
+
+            g_base_info_unref( (GIBaseInfo*) interface_info);
+        }
+        break;
+
+    case GI_TYPE_TAG_GHASH:
+        /* Possibly better to return an empty hash table? */
+        arg->v_pointer = NULL;
+        break;
+
+    case GI_TYPE_TAG_ARRAY:
+        arg->v_pointer = NULL;
+        break;
+
+    default:
+        gjs_debug(GJS_DEBUG_ERROR,
+                  "Unhandled type %s for default GArgument initialization",
+                  g_type_tag_to_string(type_tag));
+        break;
+    }
+}
+
 JSBool
 gjs_value_to_arg(JSContext  *context,
                  jsval       value,
                  GIArgInfo  *arg_info,
                  GArgument  *arg)
 {
-    GITypeInfo *type_info;
+    GITypeInfo type_info;
     gboolean result;
 
-    type_info = g_arg_info_get_type(arg_info);
+    g_arg_info_load_type(arg_info, &type_info);
 
     result =
         gjs_value_to_g_argument(context, value,
-                                type_info,
+                                &type_info,
                                 g_base_info_get_name( (GIBaseInfo*) arg_info),
                                 (g_arg_info_is_return_value(arg_info) ?
                                  GJS_ARGUMENT_RETURN_VALUE : GJS_ARGUMENT_ARGUMENT),
-                                g_arg_info_get_ownership_transfer (arg_info),
+                                g_arg_info_get_ownership_transfer(arg_info),
                                 g_arg_info_may_be_null(arg_info),
                                 arg);
-    
-    g_base_info_unref((GIBaseInfo*) type_info);
 
     return result;
 }
@@ -1205,6 +1425,95 @@ gjs_array_from_g_list (JSContext  *context,
     return result;
 }
 
+static JSBool
+gjs_array_from_g_array (JSContext  *context,
+                        jsval      *value_p,
+                        GITypeInfo *param_info,
+                        GArray     *array)
+{
+    JSObject *obj;
+    jsval elem;
+    GArgument arg;
+    JSBool result;
+    GITypeTag element_type;
+    guint i;
+
+    element_type = g_type_info_get_tag(param_info);
+    element_type = replace_gtype(element_type);
+
+    obj = JS_NewArrayObject(context, 0, JSVAL_NULL);
+    if (obj == NULL)
+      return JS_FALSE;
+
+    *value_p = OBJECT_TO_JSVAL(obj);
+
+    elem = JSVAL_VOID;
+
+#define ITERATE(type) \
+    for (i = 0; i < array->len; i++) { \
+        arg.v_##type = g_array_index(array, g##type, i); \
+        if (!gjs_value_from_g_argument(context, &elem, param_info, &arg)) \
+          goto finally; \
+        if (!JS_DefineElement(context, obj, i, elem, NULL, NULL, \
+              JSPROP_ENUMERATE)) \
+          goto finally; \
+    }
+
+    switch (element_type) {
+        case GI_TYPE_TAG_UINT8:
+          ITERATE(uint8);
+          break;
+        case GI_TYPE_TAG_INT8:
+          ITERATE(int8);
+          break;
+        case GI_TYPE_TAG_UINT16:
+          ITERATE(uint16);
+          break;
+        case GI_TYPE_TAG_INT16:
+          ITERATE(int16);
+          break;
+        case GI_TYPE_TAG_UINT32:
+          ITERATE(uint32);
+          break;
+        case GI_TYPE_TAG_INT32:
+          ITERATE(int32);
+          break;
+        case GI_TYPE_TAG_UINT64:
+          ITERATE(uint64);
+          break;
+        case GI_TYPE_TAG_INT64:
+          ITERATE(int64);
+          break;
+        case GI_TYPE_TAG_FLOAT:
+          ITERATE(float);
+          break;
+        case GI_TYPE_TAG_DOUBLE:
+          ITERATE(double);
+          break;
+        case GI_TYPE_TAG_UTF8:
+        case GI_TYPE_TAG_FILENAME:
+        case GI_TYPE_TAG_ARRAY:
+        case GI_TYPE_TAG_INTERFACE:
+        case GI_TYPE_TAG_GLIST:
+        case GI_TYPE_TAG_GSLIST:
+        case GI_TYPE_TAG_GHASH:
+          ITERATE(pointer);
+          break;
+        default:
+          gjs_throw(context, "Unknown GArray element-type %d", element_type);
+          goto finally;
+    }
+
+#undef ITERATE
+
+    result = JS_TRUE;
+
+finally:
+    JS_RemoveRoot(context, &elem);
+
+    return result;
+}
+
 static JSBool
 gjs_object_from_g_hash (JSContext  *context,
                         jsval      *value_p,
@@ -1291,8 +1600,7 @@ gjs_value_from_g_argument (JSContext  *context,
     GITypeTag type_tag;
 
     type_tag = g_type_info_get_tag( (GITypeInfo*) type_info);
-    if (type_tag != GI_TYPE_TAG_TIME_T) // we handle time_t as a non-int type
-        type_tag = normalize_int_types(type_tag);
+    type_tag = replace_gtype(type_tag);
 
     gjs_debug_marshal(GJS_DEBUG_GFUNCTION,
                       "Converting GArgument %s to jsval",
@@ -1339,11 +1647,6 @@ gjs_value_from_g_argument (JSContext  *context,
     case GI_TYPE_TAG_DOUBLE:
         return JS_NewDoubleValue(context, arg->v_double, value_p);
 
-    case GI_TYPE_TAG_TIME_T:
-        *value_p = gjs_date_from_time_t(context,
-                                        (time_t) arg->v_long);
-        return JS_TRUE;
-
     case GI_TYPE_TAG_FILENAME:
         if (arg->v_pointer)
             return gjs_string_from_filename(context, arg->v_pointer, -1, value_p);
@@ -1391,19 +1694,14 @@ gjs_value_from_g_argument (JSContext  *context,
 
                 goto out;
             } else if (interface_type == GI_INFO_TYPE_FLAGS) {
-                /* This should be fixed to work without a GType, just like Enum */
                 gtype = g_registered_type_info_get_g_type((GIRegisteredTypeInfo*)interface_info);
-                if (gtype == G_TYPE_NONE) {
-                    gjs_throw(context,
-                              "Can't yet handle flags type '%s' that is not registered with GObject",
-                              g_base_info_get_name(interface_info));
-                    goto out;
-                }
-
                 if (_gjs_flags_value_is_valid(context, gtype, arg->v_int))
                     value = INT_TO_JSVAL(arg->v_int);
 
                 goto out;
+            } else if (interface_type == GI_INFO_TYPE_STRUCT &&
+                       g_struct_info_is_foreign((GIStructInfo*)interface_info)) {
+                return gjs_struct_foreign_convert_from_g_argument(context, value_p, type_info, arg);
             }
 
             /* Everything else is a pointer type, NULL is the easy case */
@@ -1476,7 +1774,7 @@ gjs_value_from_g_argument (JSContext  *context,
     case GI_TYPE_TAG_ARRAY:
         if (arg->v_pointer == NULL) {
             /* OK, but no conversion to do */
-        } else {
+        } else if (g_type_info_get_array_type(type_info) == GI_ARRAY_TYPE_C) {
 
             if (g_type_info_is_zero_terminated(type_info)) {
                 GITypeInfo *param_info;
@@ -1488,27 +1786,13 @@ gjs_value_from_g_argument (JSContext  *context,
 
                 param_tag = g_type_info_get_tag((GITypeInfo*) param_info);
 
-                if (param_tag != GI_TYPE_TAG_UTF8) {
-                    gjs_throw(context, "FIXME: Only supporting null-terminated arrays of strings");
-                    result = JS_FALSE;
+                if (param_tag == GI_TYPE_TAG_UTF8) {
+                    result = gjs_array_from_strv(context,
+                                                 value_p,
+                                                 arg->v_pointer);
                 } else {
-                    GSList *list;
-                    char ** strv;
-
-                    list = NULL;
-                    for (strv = arg->v_pointer; *strv; strv ++) {
-                        list = g_slist_prepend(list, *strv);
-                    }
-                    list = g_slist_reverse(list);
-
-                    result = gjs_array_from_g_list(context,
-                                                   value_p,
-                                                   type_tag,
-                                                   param_info,
-                                                   NULL,
-                                                   list);
-
-                    g_slist_free(list);
+                    gjs_throw(context, "FIXME: Only supporting null-terminated arrays of strings");
+                    result = FALSE;
                 }
 
                 g_base_info_unref((GIBaseInfo*) param_info);
@@ -1518,6 +1802,31 @@ gjs_value_from_g_argument (JSContext  *context,
                 gjs_throw(context, "FIXME: Only supporting zero-terminated ARRAYs");
                 return JS_FALSE;
             }
+        } else if (g_type_info_get_array_type(type_info) == GI_ARRAY_TYPE_BYTE_ARRAY) {
+            JSObject *array = gjs_byte_array_from_byte_array(context,
+                                                             (GByteArray*)arg->v_pointer);
+            if (!array) {
+                gjs_throw(context, "Couldn't convert GByteArray to a ByteArray");
+                return JS_FALSE;
+            }
+            *value_p = OBJECT_TO_JSVAL(array);
+        } else {
+            /* this assumes the array type is one of GArray, GPtrArray or
+             * GByteArray */
+            GITypeInfo *param_info;
+            gboolean result;
+
+            param_info = g_type_info_get_param_type(type_info, 0);
+            g_assert(param_info != NULL);
+
+            result = gjs_array_from_g_array(context,
+                                            value_p,
+                                            param_info,
+                                            arg->v_pointer);
+
+            g_base_info_unref((GIBaseInfo*) param_info);
+
+            return result;
         }
         break;
 
@@ -1568,18 +1877,6 @@ gjs_value_from_g_argument (JSContext  *context,
         }
         break;
 
-    case GI_TYPE_TAG_SHORT:
-    case GI_TYPE_TAG_USHORT:
-    case GI_TYPE_TAG_INT:
-    case GI_TYPE_TAG_UINT:
-    case GI_TYPE_TAG_LONG:
-    case GI_TYPE_TAG_ULONG:
-    case GI_TYPE_TAG_SIZE:
-    case GI_TYPE_TAG_SSIZE:
-    case GI_TYPE_TAG_GTYPE:
-        /* these types are converted by normalize_int_types */
-        g_assert_not_reached();
-
     default:
         gjs_debug(GJS_DEBUG_ERROR,
                   "Unhandled type %s converting GArgument to JavaScript",
@@ -1637,7 +1934,7 @@ gjs_g_arg_release_internal(JSContext  *context,
                            GArgument  *arg)
 {
     JSBool failed;
-    
+
     g_assert(transfer != GI_TRANSFER_NOTHING);
 
     failed = JS_FALSE;
@@ -1649,21 +1946,12 @@ gjs_g_arg_release_internal(JSContext  *context,
     case GI_TYPE_TAG_UINT8:
     case GI_TYPE_TAG_INT16:
     case GI_TYPE_TAG_UINT16:
-    case GI_TYPE_TAG_SHORT:
-    case GI_TYPE_TAG_USHORT:
-    case GI_TYPE_TAG_INT:
     case GI_TYPE_TAG_INT32:
-    case GI_TYPE_TAG_UINT:
     case GI_TYPE_TAG_UINT32:
     case GI_TYPE_TAG_INT64:
     case GI_TYPE_TAG_UINT64:
-    case GI_TYPE_TAG_LONG:
-    case GI_TYPE_TAG_ULONG:
     case GI_TYPE_TAG_FLOAT:
     case GI_TYPE_TAG_DOUBLE:
-    case GI_TYPE_TAG_SSIZE:
-    case GI_TYPE_TAG_SIZE:
-    case GI_TYPE_TAG_TIME_T:
         break;
 
     case GI_TYPE_TAG_FILENAME:
@@ -1682,6 +1970,11 @@ gjs_g_arg_release_internal(JSContext  *context,
 
             interface_type = g_base_info_get_type(interface_info);
 
+            if (interface_type == GI_INFO_TYPE_STRUCT &&
+                g_struct_info_is_foreign((GIStructInfo*)interface_info))
+                return gjs_struct_foreign_release_g_argument(context,
+                        transfer, type_info, arg);
+
             if (interface_type == GI_INFO_TYPE_ENUM || interface_type == GI_INFO_TYPE_FLAGS)
                 goto out;
 
@@ -1756,15 +2049,18 @@ gjs_g_arg_release_internal(JSContext  *context,
         break;
 
     case GI_TYPE_TAG_ARRAY:
+    {
+        GIArrayType array_type = g_type_info_get_array_type(type_info);
+
         if (arg->v_pointer == NULL) {
             /* OK */
-        } else {
+        } else if (array_type == GI_ARRAY_TYPE_C) {
             GITypeInfo *param_info;
             GITypeTag element_type;
 
             param_info = g_type_info_get_param_type(type_info, 0);
             element_type = g_type_info_get_tag(param_info);
-            element_type = normalize_int_types(element_type);
+            element_type = replace_gtype(element_type);
 
             switch (element_type) {
             case GI_TYPE_TAG_UTF8:
@@ -1789,8 +2085,73 @@ gjs_g_arg_release_internal(JSContext  *context,
             }
 
             g_base_info_unref((GIBaseInfo*) param_info);
+        } else if (array_type == GI_ARRAY_TYPE_ARRAY) {
+            GITypeInfo *param_info;
+            GITypeTag element_type;
+
+            param_info = g_type_info_get_param_type(type_info, 0);
+            element_type = g_type_info_get_tag(param_info);
+            element_type = replace_gtype(element_type);
+
+            switch (element_type) {
+            case GI_TYPE_TAG_UINT8:
+            case GI_TYPE_TAG_UINT16:
+            case GI_TYPE_TAG_UINT32:
+            case GI_TYPE_TAG_UINT64:
+            case GI_TYPE_TAG_INT8:
+            case GI_TYPE_TAG_INT16:
+            case GI_TYPE_TAG_INT32:
+            case GI_TYPE_TAG_INT64:
+                g_array_free((GArray*) arg->v_pointer, TRUE);
+                break;
+
+            case GI_TYPE_TAG_UTF8:
+            case GI_TYPE_TAG_FILENAME:
+            case GI_TYPE_TAG_ARRAY:
+            case GI_TYPE_TAG_INTERFACE:
+            case GI_TYPE_TAG_GLIST:
+            case GI_TYPE_TAG_GSLIST:
+            case GI_TYPE_TAG_GHASH:
+                if (transfer == GI_TRANSFER_CONTAINER) {
+                    g_array_free((GArray*) arg->v_pointer, TRUE);
+                } else if (transfer == GI_TRANSFER_EVERYTHING) {
+                    GArray *array = arg->v_pointer;
+                    guint i;
+
+                    for (i = 0; i < array->len; i++) {
+                        GArgument arg;
+
+                        arg.v_pointer = g_array_index (array, gpointer, i);
+                        gjs_g_argument_release(context,
+                                               GI_TRANSFER_EVERYTHING,
+                                               param_info,
+                                               &arg);
+                    }
+
+                    g_array_free (array, TRUE);
+                }
+
+                break;
+
+            default:
+                gjs_throw(context,
+                          "Don't know how to release GArray element-type %d",
+                          element_type);
+                failed = JS_TRUE;
+            }
+
+            g_base_info_unref((GIBaseInfo*) param_info);
+        } else if (array_type == GI_ARRAY_TYPE_BYTE_ARRAY) {
+            if (transfer == GI_TRANSFER_CONTAINER) {
+                g_byte_array_free ((GByteArray*)arg->v_pointer, FALSE);
+            } else if (transfer == GI_TRANSFER_EVERYTHING) {
+                g_byte_array_free ((GByteArray*)arg->v_pointer, TRUE);
+            }
+        } else {
+            g_assert_not_reached();
         }
         break;
+    }
 
     case GI_TYPE_TAG_GSLIST:
         if (transfer != GI_TRANSFER_CONTAINER) {
