@@ -31,6 +31,7 @@
 #include "function.h"
 #include "value.h"
 #include "keep-alive.h"
+#include "gjs_gi_trace.h"
 
 #include <gjs/gjs.h>
 
@@ -648,8 +649,8 @@ object_instance_constructor(JSContext *context,
      * class as us, but if we're constructing an instance, the prototype
      * has the same class.
      */
-    obj_class = JS_GetClass(context, obj);
-    proto_class = JS_GetClass(context, proto);
+    obj_class = JS_GET_CLASS(context, obj);
+    proto_class = JS_GET_CLASS(context, proto);
 
     is_proto = (obj_class != proto_class);
 
@@ -765,6 +766,9 @@ object_instance_constructor(JSContext *context,
         gjs_debug_lifecycle(GJS_DEBUG_GOBJECT,
                             "JSObject created with GObject %p %s",
                             priv->gobj, g_type_name_from_instance((GTypeInstance*) priv->gobj));
+
+        TRACE(GJS_OBJECT_PROXY_NEW(priv, priv->gobj, g_base_info_get_namespace ( (GIBaseInfo*) priv->info),
+                                    g_base_info_get_name ( (GIBaseInfo*) priv->info) ));
     }
 
     return JS_TRUE;
@@ -786,6 +790,9 @@ object_instance_finalize(JSContext *context,
     if (priv == NULL)
         return; /* we are the prototype, not a real instance, so constructor never called */
 
+    TRACE(GJS_OBJECT_PROXY_FINALIZE(priv, priv->gobj, g_base_info_get_namespace ( (GIBaseInfo*) priv->info),
+                                    g_base_info_get_name ( (GIBaseInfo*) priv->info) ));
+
     if (priv->gobj) {
         g_assert(priv->gobj->ref_count > 0);
         set_js_obj(context, priv->gobj, NULL);
@@ -1098,23 +1105,23 @@ to_string_func(JSContext *context,
     JSBool ret;
     const char *namespace;
     const char *name;
-    
+
     *retval = JSVAL_VOID;
 
     priv = priv_from_js(context, obj);
 
     if (priv == NULL)
         return JS_FALSE; /* wrong class passed in */
-    
+
     namespace = g_base_info_get_namespace( (GIBaseInfo*) priv->info);
-    name = g_base_info_get_name( (GIBaseInfo*) priv->info); 
+    name = g_base_info_get_name( (GIBaseInfo*) priv->info);
 
     if (priv->gobj == NULL) {
         strval = g_strdup_printf ("[object prototype of GIName:%s.%s jsobj@%p]", namespace, name, obj);
     } else {
         strval = g_strdup_printf ("[object instance proxy GIName:%s.%s jsobj@%p native@%p]", namespace, name, obj, priv->gobj);
     }
-    
+
     ret = gjs_string_from_utf8 (context, strval, -1, retval);
     g_free (strval);
     return ret;
@@ -1371,9 +1378,9 @@ gjs_define_object_class(JSContext     *context,
                                        /* number of constructor args */
                                        0,
                                        /* props of prototype */
-                                       &gjs_object_instance_proto_props[0],
+                                       parent_proto ? NULL : &gjs_object_instance_proto_props[0],
                                        /* funcs of prototype */
-                                       &gjs_object_instance_proto_funcs[0],
+                                       parent_proto ? NULL : &gjs_object_instance_proto_funcs[0],
                                        /* props of constructor, MyConstructor.myprop */
                                        NULL,
                                        /* funcs of constructor, MyConstructor.myfunc() */
@@ -1392,7 +1399,7 @@ gjs_define_object_class(JSContext     *context,
     priv->gtype = gtype;
 
     gjs_debug(GJS_DEBUG_GOBJECT, "Defined class %s prototype %p class %p in object %p",
-              constructor_name, prototype, JS_GetClass(context, prototype), in_object);
+              constructor_name, prototype, JS_GET_CLASS(context, prototype), in_object);
 
     /* Now get the constructor we defined in
      * gjs_init_class_dynamic
@@ -1451,9 +1458,9 @@ get_obj_key(JSRuntime *runtime,
         buf[2] = '-';
         d.ptr = runtime;
         for (i = 0; i < sizeof(void*); i++) {
-        	int offset = OBJ_KEY_PREFIX_LEN+(i*2);
-        	buf[offset] = 'a' + ((d.bytes[i] & 0xf0) >> 4);
-        	buf[offset+1] = 'a' + (d.bytes[i] & 0x0f);
+                int offset = OBJ_KEY_PREFIX_LEN+(i*2);
+                buf[offset] = 'a' + ((d.bytes[i] & 0xf0) >> 4);
+                buf[offset+1] = 'a' + (d.bytes[i] & 0x0f);
         }
         buf[OBJ_KEY_LEN] = '\0';
         strcpy(cached_buf, buf);
